@@ -2,12 +2,12 @@
 
 #  ************************** Copyrights and license ***************************
 #
-# This file is part of gcovr 8.2, a parsing and reporting tool for gcov.
-# https://gcovr.com/en/8.2
+# This file is part of gcovr 8.3, a parsing and reporting tool for gcov.
+# https://gcovr.com/en/8.3
 #
 # _____________________________________________________________________________
 #
-# Copyright (c) 2013-2024 the gcovr authors
+# Copyright (c) 2013-2025 the gcovr authors
 # Copyright (c) 2013 Sandia Corporation.
 # Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 # the U.S. Government retains certain rights in this software.
@@ -22,35 +22,36 @@ import os
 import re
 import sys
 
-from argparse import ArgumentParser
+from argparse import ArgumentError, ArgumentParser, Namespace
+from typing import Any, Optional
 import traceback
-from typing import Iterable
-
-if sys.version_info >= (3, 11):
-    import tomllib
-else:
-    import tomli as tomllib
 
 from .configuration import (
-    ConfigEntry,
     argument_parser_setup,
     config_entries_from_dict,
     merge_options_and_set_defaults,
     parse_config_file,
     parse_config_into_dict,
 )
-from .utils import (
-    AlwaysMatchFilter,
-    DirectoryPrefixFilter,
+from .coverage import CoverageContainer
+from .logging import (
     configure_logging,
     update_logging,
 )
+from .filter import (
+    AlwaysMatchFilter,
+    DirectoryPrefixFilter,
+)
 from .version import __version__
-from .coverage import CovData, SummarizedStats
 
 # formats
 from . import formats as gcovr_formats
 
+if sys.version_info >= (3, 11):
+    import tomllib
+else:
+    import tomli as tomllib
+
 LOGGER = logging.getLogger("gcovr")
 
 
@@ -67,74 +68,82 @@ EXIT_WRITE_ERROR = 128
 #
 # Exits with status 2 if below threshold
 #
-def fail_under(
-    covdata: CovData,
-    threshold_line,
-    threshold_branch,
-    threshold_decision,
-    threshold_function,
-):
-    stats = SummarizedStats.from_covdata(covdata)
-
-    line_nok = False
-    if threshold_line > 0.0:
-        # If there are no lines, mark as uncovered
-        # (indicates no data at all, likely an error).
-        percent_lines = stats.line.percent_or(0.0)
-
-        if percent_lines < threshold_line:
-            line_nok = True
-            LOGGER.error(
-                f"failed minimum line coverage (got {percent_lines}%, minimum {threshold_line}%)"
-            )
-
-    branch_nok = False
-    if threshold_branch > 0.0:
-        # Allow data with no branches.
-        percent_branches = stats.branch.percent_or(100.0)
-        if percent_branches < threshold_branch:
-            branch_nok = True
-            LOGGER.error(
-                f"failed minimum branch coverage (got {percent_branches}%, minimum {threshold_branch}%)"
-            )
-
-    decision_nok = False
-    if threshold_decision > 0.0:
-        # Allow data with no decisions.
-        percent_decision = stats.decision.percent_or(100.0)
-        if percent_decision < threshold_decision:
-            decision_nok = True
-            LOGGER.error(
-                f"failed minimum decision coverage (got {percent_decision}%, minimum {threshold_decision}%)"
-            )
-
-    function_nok = False
-    if threshold_function > 0.0:
-        # Allow data with no functions.
-        percent_function = stats.function.percent_or(100.0)
-        if percent_function < threshold_function:
-            function_nok = True
-            LOGGER.error(
-                f"failed minimum function coverage (got {percent_function}%, minimum {threshold_function}%)"
-            )
-
+def get_exit_code(
+    covdata: CoverageContainer,
+    threshold_line: float,
+    threshold_branch: float,
+    threshold_decision: float,
+    threshold_function: float,
+) -> int:
+    """Fail depending on the coverage result."""
     exit_code = 0
-    if line_nok:
-        exit_code |= EXIT_LINE_NOK
-    if branch_nok:
-        exit_code |= EXIT_BRANCH_NOK
-    if decision_nok:
-        exit_code |= EXIT_DECISION_NOK
-    if function_nok:
-        exit_code |= EXIT_FUNCTION_NOK
-    if exit_code != 0:
-        sys.exit(exit_code)
+
+    if (
+        threshold_line > 0.0
+        or threshold_branch > 0.0
+        or threshold_decision > 0.0
+        or threshold_function > 0.0
+    ):
+        stats = covdata.stats
+
+        line_nok = False
+        if threshold_line > 0.0:
+            # If there are no lines, mark as uncovered
+            # (indicates no data at all, likely an error).
+            percent_lines = stats.line.percent_or(0.0)
+
+            if percent_lines < threshold_line:
+                line_nok = True
+                LOGGER.error(
+                    f"failed minimum line coverage (got {percent_lines}%, minimum {threshold_line}%)"
+                )
+
+        branch_nok = False
+        if threshold_branch > 0.0:
+            # Allow data with no branches.
+            percent_branches = stats.branch.percent_or(100.0)
+            if percent_branches < threshold_branch:
+                branch_nok = True
+                LOGGER.error(
+                    f"failed minimum branch coverage (got {percent_branches}%, minimum {threshold_branch}%)"
+                )
+
+        decision_nok = False
+        if threshold_decision > 0.0:
+            # Allow data with no decisions.
+            percent_decision = stats.decision.percent_or(100.0)
+            if percent_decision < threshold_decision:
+                decision_nok = True
+                LOGGER.error(
+                    f"failed minimum decision coverage (got {percent_decision}%, minimum {threshold_decision}%)"
+                )
+
+        function_nok = False
+        if threshold_function > 0.0:
+            # Allow data with no functions.
+            percent_function = stats.function.percent_or(100.0)
+            if percent_function < threshold_function:
+                function_nok = True
+                LOGGER.error(
+                    f"failed minimum function coverage (got {percent_function}%, minimum {threshold_function}%)"
+                )
+
+        if line_nok:
+            exit_code |= EXIT_LINE_NOK
+        if branch_nok:
+            exit_code |= EXIT_BRANCH_NOK
+        if decision_nok:
+            exit_code |= EXIT_DECISION_NOK
+        if function_nok:
+            exit_code |= EXIT_FUNCTION_NOK
+
+    return exit_code
 
 
-def create_argument_parser():
+def create_argument_parser() -> ArgumentParser:
     """Create the argument parser."""
 
-    parser = ArgumentParser(add_help=False)
+    parser = ArgumentParser(add_help=False, exit_on_error=False)
     parser.usage = "gcovr [options] [search_paths...]"
     parser.description = (
         "A utility to run gcov and summarize the coverage in simple reports."
@@ -160,18 +169,17 @@ def create_argument_parser():
 
 
 COPYRIGHT = (
-    "Copyright (c) 2013-2024 the gcovr authors\n"
+    "Copyright (c) 2013-2025 the gcovr authors\n"
     "Copyright (c) 2013 Sandia Corporation.\n"
     "Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,\n"
     "the U.S. Government retains certain rights in this software.\n"
 )
 
 
-def find_config_name(root: str, filename: str):
+def find_config_name(root: str, filename: str) -> Optional[str]:
+    """Find the configuration to use."""
     if root:
         filename = os.path.join(root, filename)
-    else:
-        filename = filename
 
     if os.path.isfile(filename):
         return filename
@@ -179,7 +187,7 @@ def find_config_name(root: str, filename: str):
     return None
 
 
-def load_config(partial_options) -> Iterable[ConfigEntry]:
+def load_config(partial_options: Namespace) -> dict[str, Any]:
     """Load a config file if configured or found by default names"""
     filename = getattr(partial_options, "config", None)
     if filename is not None:
@@ -207,14 +215,23 @@ def load_config(partial_options) -> Iterable[ConfigEntry]:
     return {}
 
 
-def main(args=None):
+def main(args: Optional[list[str]] = None) -> int:  # pylint: disable=too-many-return-statements
+    """The main entry point of GCOVR."""
     configure_logging()
-    parser = create_argument_parser()
-    cli_options = parser.parse_args(args=args)
+    try:
+        parser = create_argument_parser()
+        cli_options = parser.parse_args(args=args)
+    except SystemExit as e:
+        if e.code != 0:
+            raise AssertionError("Sanity check failed, exitcode must be 0.") from e
+        return EXIT_SUCCESS
+    except ArgumentError as e:
+        sys.stderr.write(f"gcovr: error: {e}\n")
+        return EXIT_CMDLINE_ERROR
 
     if cli_options.version:
         sys.stdout.write(f"gcovr {__version__}\n\n{COPYRIGHT}")
-        sys.exit(EXIT_SUCCESS)
+        return EXIT_SUCCESS
 
     # load the config
     cfg_options = load_config(cli_options)
@@ -230,11 +247,11 @@ def main(args=None):
         LOGGER.error(
             "the options --sort-branches without '--sort uncovered-number' or '--sort uncovered-percent' doesn't make sense."
         )
-        sys.exit(EXIT_CMDLINE_ERROR)
+        return EXIT_CMDLINE_ERROR
 
     if options.html_title == "":
         LOGGER.error("an empty --html-title= is not allowed.")
-        sys.exit(EXIT_CMDLINE_ERROR)
+        return EXIT_CMDLINE_ERROR
 
     for postfix in ["", "line", "branch"]:
         key_medium = "html_medium_threshold"
@@ -247,7 +264,7 @@ def main(args=None):
 
         if getattr(options, key_medium) == 0:
             LOGGER.error(f"value of {option_medium}= should not be zero.")
-            sys.exit(EXIT_CMDLINE_ERROR)
+            return EXIT_CMDLINE_ERROR
 
         # Inherit the defaults from the global coverage values if not set
         if postfix:
@@ -273,13 +290,13 @@ def main(args=None):
                 f"value of {option_medium}={getattr(options, key_medium)} should be\n"
                 f"lower than or equal to the value of {option_high}={getattr(options, key_high)}."
             )
-            sys.exit(EXIT_CMDLINE_ERROR)
+            return EXIT_CMDLINE_ERROR
 
     try:
         gcovr_formats.validate_options(options)
     except RuntimeError as exc:
         LOGGER.error(str(exc))
-        sys.exit(EXIT_CMDLINE_ERROR)
+        return EXIT_CMDLINE_ERROR
 
     options.starting_dir = os.path.abspath(os.getcwd())
     options.root_dir = os.path.abspath(options.root)
@@ -292,21 +309,21 @@ def main(args=None):
         # but is used to turn absolute paths into relative paths
         options.root_filter = re.compile("^" + re.escape(options.root_dir + os.sep))
 
+        options.filter = [f.build_filter() for f in options.filter]
+        if not options.filter:
+            options.filter = [DirectoryPrefixFilter(options.root_dir)]
+        options.exclude = [f.build_filter() for f in options.exclude]
+        options.include = [f.build_filter() for f in options.include]
+
+        options.gcov_filter = [f.build_filter() for f in options.gcov_filter]
+        if not options.gcov_filter:
+            options.gcov_filter = [AlwaysMatchFilter()]
+        options.gcov_exclude = [f.build_filter() for f in options.gcov_exclude]
         if options.gcov_exclude_dirs:
             options.gcov_exclude_dirs = [
                 f.build_filter() for f in options.gcov_exclude_dirs
             ]
 
-        options.exclude = [f.build_filter() for f in options.exclude]
-        options.filter = [f.build_filter() for f in options.filter]
-        if not options.filter:
-            options.filter = [DirectoryPrefixFilter(options.root_dir)]
-
-        options.gcov_exclude = [f.build_filter() for f in options.gcov_exclude]
-        options.gcov_filter = [f.build_filter() for f in options.gcov_filter]
-        if not options.gcov_filter:
-            options.gcov_filter = [AlwaysMatchFilter()]
-
         options.exclude_functions = [
             (re.compile(f[1:-1] if f[0] == "/" and f[-1] == "/" else re.escape(f)))
             for f in options.exclude_functions
@@ -316,6 +333,7 @@ def main(args=None):
             ("--root", [options.root_filter]),
             ("--filter", options.filter),
             ("--exclude", options.exclude),
+            ("--include", options.include),
             ("--gcov-filter", options.gcov_filter),
             ("--gcov-exclude", options.gcov_exclude),
             ("--gcov-exclude-directories", options.gcov_exclude_dirs),
@@ -326,18 +344,20 @@ def main(args=None):
                 LOGGER.debug(f" - {f}")
 
     except re.error as e:
-        LOGGER.error(f"Error setting up filter '{e.pattern}': {e}")
-        sys.exit(EXIT_CMDLINE_ERROR)
+        LOGGER.error(f"Error setting up filter '{e.pattern!s}': {e}")
+        return EXIT_CMDLINE_ERROR
 
     if options.exclude_lines_by_pattern:
         try:
             re.compile(options.exclude_lines_by_pattern)
         except re.error as e:
+            # mypy is thinking that the pattern can be a byte string therefore we need to explicit use !s.
+            # See also discussion https://github.com/gcovr/gcovr/pull/1028#discussion_r1855437452
             LOGGER.error(
                 "--exclude-lines-by-pattern: "
-                f"Invalid regular expression: {repr(options.exclude_lines_by_pattern)}, error: {e}"
+                f"Invalid regular expression: {options.exclude_lines_by_pattern!s}, error: {e}"
             )
-            sys.exit(EXIT_CMDLINE_ERROR)
+            return EXIT_CMDLINE_ERROR
 
     if options.exclude_branches_by_pattern:
         try:
@@ -347,44 +367,42 @@ def main(args=None):
                 "--exclude-branches-by-pattern: "
                 f"Invalid regular expression: {repr(options.exclude_branches_by_pattern)}, error: {e}"
             )
-            sys.exit(EXIT_CMDLINE_ERROR)
+            return EXIT_CMDLINE_ERROR
 
     if options.fail_under_decision > 0.0 and not options.show_decision:
         LOGGER.error("--fail-under-decision need also option --decision.")
-        sys.exit(EXIT_CMDLINE_ERROR)
+        return EXIT_CMDLINE_ERROR
+
+    if options.show_decision:
+        LOGGER.info(
+            "Attention, the decision analysis is experimental. "
+            "It uses a fragile heuristic and depends on the code format."
+        )
 
     LOGGER.info("Reading coverage data...")
     try:
-        covdata: CovData = gcovr_formats.read_reports(options)
-    except Exception as e:
-        LOGGER.error(
-            f"Error occurred while reading reports:\n{traceback.format_exc()}\n{str(e)}"
-        )
-        sys.exit(EXIT_READ_ERROR)
+        covdata = gcovr_formats.read_reports(options)
+    except Exception:  # pylint: disable=broad-exception-caught
+        LOGGER.error(f"Error occurred while reading reports:\n{traceback.format_exc()}")
+        return EXIT_READ_ERROR
 
     LOGGER.info("Writing coverage report...")
     try:
         gcovr_formats.write_reports(covdata, options)
-    except Exception as e:
+    except Exception:  # pylint: disable=broad-exception-caught
         LOGGER.error(
-            f"Error occurred while printing reports:\n{traceback.format_exc()}\n{str(e)}"
+            f"Error occurred while printing reports:\n{traceback.format_exc()}"
         )
-        sys.exit(EXIT_WRITE_ERROR)
+        return EXIT_WRITE_ERROR
 
-    if (
-        options.fail_under_line > 0.0
-        or options.fail_under_branch > 0.0
-        or options.fail_under_decision > 0.0
-        or options.fail_under_function > 0.0
-    ):
-        fail_under(
-            covdata,
-            options.fail_under_line,
-            options.fail_under_branch,
-            options.fail_under_decision,
-            options.fail_under_function,
-        )
+    return get_exit_code(
+        covdata,
+        options.fail_under_line,
+        options.fail_under_branch,
+        options.fail_under_decision,
+        options.fail_under_function,
+    )
 
 
 if __name__ == "__main__":
-    main()
+    sys.exit(main())
